@@ -1,7 +1,7 @@
 package Text::Fragment;
 
-our $DATE = '2015-09-06'; # DATE
-our $VERSION = '0.08'; # VERSION
+our $DATE = '2016-01-29'; # DATE
+our $VERSION = '0.09'; # VERSION
 
 use 5.010001;
 use strict;
@@ -24,6 +24,44 @@ our $re_id = qr/\A[A-Za-z0-9_.,:-]+\z/;
 
 our %SPEC;
 
+sub _format_quoted {
+    my $unquoted = shift;
+    my $res = "";
+    my $i = -1;
+    while (++$i < length($unquoted)) {
+        my $c = substr($unquoted, $i, 1);
+        if ($c eq '\\' or $c eq '"') {
+            $res .= "\\$c";
+        } elsif ($c !~ /[\x20-\x7F]/) {
+            # strip non-printables
+        } else {
+            $res .= $c;
+        }
+    }
+    qq("$res");
+}
+
+sub _parse_quoted {
+    my $quoted = shift;
+    $quoted =~ s/\A"//; $quoted =~ s/"\z//;
+    my $res = "";
+    my $i = -1;
+    while (++$i < length($quoted)) {
+        my $c = substr($quoted, $i, 1);
+        if ($c eq '\\') {
+            $res .= substr($quoted, ++$i, 1);
+        } else {
+            $res .= $c;
+        }
+    }
+    $res;
+}
+
+sub _format_attr_value {
+    my $val = shift;
+    $val =~ /\s|"|[^\x20-\x7f]/ ? _format_quoted($val) : $val;
+}
+
 sub _label {
     my %args  = @_;
     my $id            = $args{id} // "";
@@ -31,14 +69,16 @@ sub _label {
     my $comment_style = $args{comment_style};
     my $attrs         = $args{attrs};
 
+    my $quoted_re = qr/"(?:[^\n\r"\\]|\\[^\n\r])*"/;
+
     my $a_re;  # regex to match attributes
     my $ai_re; # also match attributes, but attribute id must be present
     if (length $id) {
-        $ai_re = qr/(?:\w+=\S*[ \t]+)*id=(?<id>\Q$id\E)(?:[ \t]+\w+=\S+)*/;
+        $ai_re = qr/(?:\w+=\S*[ \t]+)*id=(?<id>\Q$id\E)(?:[ \t]+\w+=(?:$quoted_re|\S+))*/;
     } else {
-        $ai_re = qr/(?:\w+=\S*[ \t]+)*id=(?<id>\S*)(?:[ \t]+\w+=\S+)*/;
+        $ai_re = qr/(?:\w+=\S*[ \t]+)*id=(?<id>\S*)(?:[ \t]+\w+=(?:$quoted_re|\S+))*/;
     }
-    $a_re  = qr/(?:\w+=\S*)?(?:[ \t]+\w+=\S*)*/;
+    $a_re  = qr/(?:\w+=\S*)?(?:[ \t]+\w+=(?:$quoted_re|\S+))*/;
 
     my ($ts, $te); # tag start and end
     if ($comment_style eq 'shell') {
@@ -81,9 +121,8 @@ sub _label {
     my $parse_attrs = sub {
         my $s = shift // "";
         my %a;
-        for my $a (split /[ \t]+/, $s) {
-            my ($n, $v) = split /=/, $a, 2;
-            $a{$n} = $v;
+        while ($s =~ /(\w+)=(?:($quoted_re)|(\S+))(?:\s+|\z)/g) {
+            $a{$1} = $2 ? _parse_quoted($2) : $3;
         }
         \%a;
     };
@@ -99,12 +138,13 @@ sub _label {
             my $as = "";
             if (ref($f{attrs})) {
                 for (sort keys %{ $f{attrs} }) {
-                    $as .= " " . "$_=$f{attrs}{$_}";
+                    $as .= " " . "$_="._format_attr_value($f{attrs}{$_});
                 }
             } else {
                 my $a = $parse_attrs->($f{attrs});
-                $as = join("", map {" $_=$a->{$_}"} grep {$_ ne 'id'}
-                    sort keys %$a);
+                $as = join("", map {" $_="._format_attr_value($a->{$_})}
+                               grep {$_ ne 'id'}
+                               sort keys %$a);
             }
 
             my $pl = $f{payload};
@@ -161,8 +201,6 @@ sub _doit {
                 return [400, "Undefined value for attribute name '$_'"];
             }
         }
-        $attrs->{$_} =~ /\s/s and return
-            [400,"Invalid value in attribute '$_', no whitespaces please"];
     }
 
     my $good_pattern       = $args{good_pattern};
@@ -654,7 +692,7 @@ Text::Fragment - Manipulate fragments in text
 
 =head1 VERSION
 
-This document describes version 0.08 of Text::Fragment (from Perl distribution Text-Fragment), released on 2015-09-06.
+This document describes version 0.09 of Text::Fragment (from Perl distribution Text-Fragment), released on 2016-01-29.
 
 =head1 SYNOPSIS
 
@@ -710,33 +748,6 @@ To set fragment attributes:
 
  $res = se_fragment_attrs(text=>$text, id=>'bar', attrs=>{name=>'val', ...});
 
-=head1 HOMEPAGE
-
-Please visit the project's homepage at L<https://metacpan.org/release/Text-Fragment>.
-
-=head1 SOURCE
-
-Source repository is at L<https://github.com/sharyanto/perl-Text-Fragment>.
-
-=head1 BUGS
-
-Please report any bugs or feature requests on the bugtracker website L<https://rt.cpan.org/Public/Dist/Display.html?Name=Text-Fragment>
-
-When submitting a bug or request, please include a test-file or a
-patch to an existing test-file that illustrates the bug or desired
-feature.
-
-=head1 AUTHOR
-
-perlancar <perlancar@cpan.org>
-
-=head1 COPYRIGHT AND LICENSE
-
-This software is copyright (c) 2015 by perlancar@cpan.org.
-
-This is free software; you can redistribute it and/or modify it under
-the same terms as the Perl 5 programming language system itself.
-
 =head1 DESCRIPTION
 
 
@@ -807,6 +818,8 @@ Newline deletion behaviour: if fragment at the bottom of text does not end with
 newline (which is considered bad style), the text after the fragment is deleted
 will also not end with newline.
 
+This function is not exported by default, but exportable.
+
 Arguments ('*' denotes required arguments):
 
 =over 4
@@ -858,6 +871,8 @@ Get fragment with a certain ID in text.
 
 If there are multiple occurences of the fragment with the same ID ,
 
+This function is not exported by default, but exportable.
+
 Arguments ('*' denotes required arguments):
 
 =over 4
@@ -908,6 +923,8 @@ not end with newline (which is considered bad style), the inserted fragment will
 also not end with newline. Except when original text is an empty string, in
 which case an ending newline will still be added.
 
+This function is not exported by default, but exportable.
+
 Arguments ('*' denotes required arguments):
 
 =over 4
@@ -985,6 +1002,8 @@ same payload that needs to be inserted already exists in the text).
 
 List fragments in text.
 
+This function is not exported by default, but exportable.
+
 Arguments ('*' denotes required arguments):
 
 =over 4
@@ -1027,6 +1046,8 @@ Set/unset attributes of a fragment.
 
 If there are multiple occurences of the fragment with the same ID ,
 
+This function is not exported by default, but exportable.
+
 Arguments ('*' denotes required arguments):
 
 =over 4
@@ -1071,4 +1092,31 @@ attributes before being modified).
 
 Return 404 if fragment is not found.
 
+=head1 HOMEPAGE
+
+Please visit the project's homepage at L<https://metacpan.org/release/Text-Fragment>.
+
+=head1 SOURCE
+
+Source repository is at L<https://github.com/sharyanto/perl-Text-Fragment>.
+
+=head1 BUGS
+
+Please report any bugs or feature requests on the bugtracker website L<https://rt.cpan.org/Public/Dist/Display.html?Name=Text-Fragment>
+
+When submitting a bug or request, please include a test-file or a
+patch to an existing test-file that illustrates the bug or desired
+feature.
+
+=head1 AUTHOR
+
+perlancar <perlancar@cpan.org>
+
+=head1 COPYRIGHT AND LICENSE
+
+This software is copyright (c) 2016 by perlancar@cpan.org.
+
+This is free software; you can redistribute it and/or modify it under
+the same terms as the Perl 5 programming language system itself.
+
 =cut
